#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#else
#include <strings.h>
#endif

#include <stdio.h>
#include "compilerApi.h"
#include "strategy.h"

#include "anls_att.h"

#define INDEX_EXP  1000

//extern "C" int strcasecmp(...);

int nindo_level=0;
int indo_level=0;
int sec_number;

extern Flag in_independent;

extern int element_intrinsic_call;

extern int getreflevel(SgSymbol * index, SgExpression * expr);
extern SgExpression * getlinerexp(int i, SgSymbol * symb, SgExpression * expr);
extern void delete_dummy_array(SgStatement *);

extern void prepare_actualdad_for_proc(SgStatement *,SgExpression *,int &);
extern void change_procstmt(SgStatement *);
extern void change_prochedr(SgStatement *);

extern void trans_if_stmt(SgStatement *);
extern void trans_assign_reference(SgAssignStmt * , int) ;
extern void trans_funccall(SgStatement * ,SgExpression *);
extern void wrap_assign_on(SgStatement *);

extern void  analyse_collapse(SgStatement *,SgArrayRefExp *);
extern void collect_array_ref(SgStatement *, SgExpression *, int);
extern void analyse_array_ref(SgStatement *,SgExpression *,SgStatement *,SgSymbol *);
extern SgExpression *get_first_address(SgStatement *s,SgArrayRefExp * arrayexp);

extern void get_scalar_stmt(SgStatement *,SgStatement *);
extern void get_scalar_exp(SgStatement *,SgExpression *);


/*================================================================
Independent_header_def class member function
================================================================*/
Independent_header_def::Independent_header_def(void)
{
  number=0;
  for (int i=0; i<MAX_INDEX_DIM; i++)
    for (int j=0; j<3; j++)
      index[i].indextriple[j]=NULL;	//simplified for now
}

int Independent_header_def::getindexnumber(void)
{
  return number;
}

void Independent_header_def::setindexsymb(SgSymbol * s)
{
  number++;
  index[number-1].indexvar=s;
}

SgSymbol * Independent_header_def::getindexsymb(int i)
{
  return index[i].indexvar;
}

void Independent_header_def::setindextriple(int i, SgExpression * e)
{
  index[number-1].indextriple[i]=e;
}

SgExpression * Independent_header_def::getindextriple(int i, int j)
{
  return index[i].indextriple[j];
}

/*================================================================
independent_reference class member function
================================================================*/

//construction func, same as forall_reference
independent_reference::independent_reference()
{
   for (int i=0; i<MAX_ARRAY_DIM; i++)
    for (int j=0; j<MAX_INDEX_DIM; j++)
    {
      index_level[i][j]=0;
      stride[i][j]=NULL;
      offset[i][j]=NULL;
    }

  intent = as_rhs;
  status = NONE;
  sameAs = -1;
  left_count = 0;
  right_count = 0;

}

void independent_reference::setIntent(Usedfor it)
{
	intent = it;
}

int independent_reference::getIntent()
{
	return intent;
} 

void independent_reference::setLeftCount(int cnt)
{
	left_count = cnt;
}

int independent_reference::getLeftCount()
{
	return left_count;
} 

void independent_reference::setRightCount(int cnt)
{
	right_count = cnt;
}

int independent_reference::getRightCount()
{
	return right_count;
} 

void independent_reference::setExpr(SgExpression * e)
{
  expr = e;
}

SgExpression * independent_reference::getexpression()
{
  return expr;
}

void independent_reference::setStmt(SgStatement * stmt)
{
  used = stmt;
}

SgStatement * independent_reference::getStmt()
{
  return used;
}

void independent_reference::setIndex(int arraydim, int indexdim, int IL) 
{
  index_level[arraydim][indexdim]=IL;
}

int independent_reference::getIndex(int arraydim, int indexdim)
{
  return index_level[arraydim][indexdim];
}

void independent_reference::setstride(int arraydim, int indexdim, SgExpression * expr)
{
  stride[arraydim][indexdim]=expr;
}

SgExpression * independent_reference::getstride(int arraydim,int indexdim)
{
  return stride[arraydim][indexdim];
}

void independent_reference::setoffset(int arraydim, int indexdim, SgExpression * expr)
{
  offset[arraydim][indexdim]=expr;
}

SgExpression * independent_reference::getoffset(int arraydim,int indexdim)
{
  return offset[arraydim][indexdim];
}

void independent_reference::setCommunication(Communication c)
{
  status=c;
}

Communication independent_reference::getCommunication()
{
  return status;
}

int independent_reference::refered(int index)
{
   int i;
   for (i=0; i<MAX_ARRAY_DIM; i++)
        if (index_level[i][index]>0) break;
   if (i==MAX_ARRAY_DIM) return 0;
   return 1;
}

int independent_reference::isScalar()
{
  for (int i=0; i<MAX_ARRAY_DIM; i++)
    for (int j=0; j<MAX_INDEX_DIM; j++)
    {
      if (index_level[i][j]!=0) return 0;
    }
  return 1;
}

int independent_reference::indexNumberOnDim(int dim)
{
  for (int i=0; i<MAX_INDEX_DIM; i++)
    if (index_level[dim][i]>0)
      return i;
  return -1;
}

int independent_reference::dimNumberOnIndex(int dim)
{
  for (int i=0; i<MAX_ARRAY_DIM; i++)
    if (index_level[i][dim]>0)
      return i;
  return -1;
}

int independent_reference::hasScalarDim()
{
  SgArrayRefExp *array = isSgArrayRefExp(expr);
  ASSERT(array);
  int rank = array->numberOfSubscripts();
  for (int i=0; i<rank; i++)
        if (indexNumberOnDim(i) == -1) return 1;
  return 0;
}

/*================================================================
general purpose tools
================================================================*/
//find lhs reference of a stmt inside indo
independent_reference * get_lhsref(SgStatement * s)
{
	independent_reference * lhsref;
        int has_lhs=0;

	//assuming only one lhsref!
        for (int i=0; i<s->numberOfAttributes(Independent_Ref); i++) {
            lhsref=(independent_reference *) ((s->getAttribute(i, Independent_Ref))->getAttributeData());
            if(lhsref->getIntent()==as_lhs)
		{ has_lhs=1; break;}
        };

        if(!has_lhs) {
            printf("Error: Statement has no lhs:%d\n",s->id());
            return NULL;
        };

        return lhsref;

}

//see if the rhs array is collapsed, if it is not dummy and collapsed, then 1
int find_collapse(SgStatement * s,independent_reference * rhs)
{
   	SgSymbol *symb = rhs->getexpression()->symbol();

     	if (array_all_collapsed(symb) == 1)
        {
            if (is_dummy(s, symb) == 0 )
                return 1;
            else {
                        if (symb->numberOfAttributes(ALL_COLLAPSED) != 0)
                                return 1;
                 }
        }

        return 0;
}

// if no need communication,return 1
// need future work on this!
int detect_indo_no_communication(SgStatement * s, independent_reference * rhs, int assume_none)
{
	//if same as lhs, no comm(to be further detected)
   	if(rhs->sameAs==0) return 1;	

	//see if collapsed
     	if (find_collapse(s,rhs)== 1) return 1;

	//temporarily put here, todo real comm_detect in the future
	if(assume_none){
		//printf("ASSUMING NONE COMMUNICATION IN INDEP_DO.\n");
		return 1;
	}
	
        return 0;
}

//local index transform to global index
void change_indo_scalar_reference(SgStatement * s, independent_reference *
 indref)
{
  Independent_header_def * indohdef = (Independent_header_def *) s->getAttribute(0, INDEPHDER_Def)->getAttributeData();

  SgExpression * oldscalar = indref->getexpression();
  SgSymbol * scalarsym = oldscalar->symbol();

  for(int i=0; i<indohdef->getindexnumber(); i++){
    if(indohdef->getindexsymb(i)->id()==scalarsym->id()) {

        SgSymbol * glb = s->getSymbolInCurrentScope(mixstring, "glb", scalarsym->identifier());
        SgExpression * glbexp = new SgExpression (VAR_REF, NULL, NULL, glb);

        oldscalar->replace(*glbexp);

        break;
    }
  }//end for

}

//find if we already have such an expression inserted
//return: 0--same as lhs; -1--not same as anything; i--same as rhs no.i
int indo_find_same_as(SgStatement *s , SgExpression *expr)
{
 for (int i=0; i<s->numberOfAttributes(Independent_Ref); i++) {
        independent_reference * indoref = (independent_reference *) (s->attributeValue(i,Independent_Ref));

        if ( (*indoref->getexpression()) == *expr )
        {
		if ( indoref->getIntent() == as_rhs)
		{
			int sameasi = indoref->getRightCount();
			if (!silentFlag)
			printf("same as :%d\n",sameasi);
                	return sameasi;
        	}
		else return 0;
	}
  }

 return -1;
}

//transform forall into serialized do loop, inside indo
//because no communication should exist inside indo
void forall_do(SgStatement *s)
{
        SgForallStmt *forall = isSgForallStmt(s);
        ASSERT(forall);
//printf("Serialize Forall Statement to Do-loop.\n");
        SgStatement *body = s->childList1(0);
        SgStatement *do_body = body->copyPtr();
        SgExpression *mask = forall->getMask();
        SgStatement *newdo;
        if (mask)
                newdo = new SgLogIfStmt(*mask, *do_body);
        else
                newdo = do_body;
        for (int i=0; i<forall->indexNumber(); i++)
        {
                SgExpression *index_op = forall->getIndex(i);
                SgSymbol *index = forall->getIndexName(i);
                SgTripletExp *ddot = (SgTripletExp*)index_op->rhs();
                ASSERT(ddot);
                SgExpression *start = ddot->item(0);
                SgExpression *end = ddot->item(1);
                SgExpression *stride = ddot->item(2);
                newdo = new SgForStmt(index, start, end, stride, newdo);

        }
        s->insertStmtAfter(*newdo, *s->controlParent());
        s->deleteStmt();
}


/*==================================================================
optimization related work
==================================================================*/

//judge if array argument's section creation can be pulled out of indo
//if this section is not to be used inside subroutine, then yes
int optimize_indo_call(SgStatement * s)
{
  SgCallStmt *call = isSgCallStmt(call);
  ASSERT(call);
  independent_reference * ind_ref = get_lhsref(s);
  SgExpression *exp = ind_ref->getexpression();
  SgSymbol *symb = exp->symbol();
  ASSERT(symb);
  int storage = getArrayStorage(symb);

  if (storage == COLUMN_MAJOR) {
	if (array_dim_collapsed(symb, 0) == 1) return 1;
  }
  else {
	SgArrayType *type = isSgArrayType(symb->type());
	ASSERT(type);
	int rank = type->dimension();
	if (array_dim_collapsed(symb, rank-1) == 1) return 1;
  }
  return 0;
}

//optimize pcrc_glob(), by pull it out of indo, and calculate through 
//local first address and step
void indo_add_init_value(SgStatement *s,SgStatement *loop,SgSymbol *scalarsym,SgSymbol *irg, int lc)
{
        SgSymbol * step0 = s->getSymbolInCurrentScope(mixstring, "step", scalarsym->identifier(),0);

        SgSymbol * lil = s->getSymbolInCurrentScope(mixstring, "lil", scalarsym->identifier(), lc);
        SgSymbol * lis = s->getSymbolInCurrentScope(mixstring, "lis", scalarsym->identifier(), lc);

        SgSymbol * step = s->getSymbolInCurrentScope(mixstring, "step", scalarsym->identifier());

        {
        SgExpression *lilexp = new SgVarRefExp(lil);
        SgExpression *globexp = make_glob(s,irg,lilexp);
        SgExpression *step0exp = new SgVarRefExp(step0);

        // add pcrc_step_i0 = pcrc_glob(pcrc_irg0 , pcrc_lil_i)
        SgAssignStmt *ass = new SgAssignStmt(*step0exp , *globexp);
        loop->insertStmtBefore(*ass, *loop->controlParent());
        }

        {
        SgExpression *stepexp = new SgVarRefExp(step);
        SgExpression *lilexp = new SgVarRefExp(lil);
        SgExpression *lisexp = new SgVarRefExp(lis);
        SgExpression *exp = &(*lisexp  + *lilexp );
        SgExpression *globexp1 = make_glob(s,irg,exp);
        SgExpression *globexp2 = make_glob(s,irg,lilexp);
        exp = &( *globexp1 - *globexp2 );
        SgAssignStmt *ass = new SgAssignStmt(*stepexp , *exp);
        SgStatement *do_loop = loop;
        while ( do_loop->variant() == FOR_NODE  && do_loop->controlParent()
 &&
                do_loop->controlParent()->variant() == FOR_NODE )
        {
                do_loop=do_loop->controlParent();
        }
        do_loop->insertStmtBefore(*ass, *do_loop->controlParent());
        }

        // add increase
        {
        SgExpression *step0exp = new SgExpression(VAR_REF,NULL,NULL,step0);
        SgExpression *stepexp = new SgExpression(VAR_REF,NULL,NULL,step);
        SgExpression *exp = &(*step0exp + *stepexp );
        SgAssignStmt *ass = new SgAssignStmt(*step0exp->copyPtr() , *exp);
        SgStatement *laststmt = loop->childList1(loop->numberOfChildrenList1()-3);              // enddo stmt , control_end
        ASSERT(laststmt);
        laststmt->insertStmtAfter(*ass, *laststmt->controlParent());
        }
}

//replace every pcrc_glob() inside indo with the calculated variable
void indo_optimize_glob_index(SgStatement *outer, SgStatement *inner, Independent_header_def *indohdef, int lc)
{
  int i,j;

  outer->deleteAllAttribute(INDEX_EXP);
  get_scalar_stmt(outer,outer);

  for (i=0; i<indohdef->getindexnumber(); i++) {
     int add=0;
     for (j=0; j<outer->numberOfAttributes(INDEX_EXP); j++) {
        SgExpression *exp = (SgExpression *)(outer->attributeValue(j,INDEX_EXP));
        SgExpression *index=exp->lhs()->rhs()->lhs()->lhs()->rhs();
        SgSymbol *irg=exp->lhs()->lhs()->symbol();
        SgSymbol *scalarsym = index->symbol();

        if(strcasecmp(indohdef->getindexsymb(i)->identifier(),scalarsym->identifier()) == 0)
        {
                add=1;
                // find the loop of index
                SgStatement *do_loop = inner;
                while ( do_loop->variant() == FOR_NODE )
                {
                        if ( strcasecmp( ((SgForStmt *)do_loop)->doName()->identifier() ,scalarsym->identifier()) == 0 )
                        break;
                        do_loop=do_loop->controlParent();
                }
                // add init value of pcrc_step_i0 and pcrc_step_i
//printf("optimize local=>global:%s\n",scalarsym->identifier());
                ASSERT(do_loop);
                indo_add_init_value(outer,do_loop,scalarsym,irg,lc);
        }
        if (add == 1) break;
     }
  }

  // change  pcrc_glob => pcrc_step_i0
  for (i=0; i<outer->numberOfAttributes(INDEX_EXP); i++) {
        SgExpression *exp = (SgExpression *)(outer->attributeValue(i,INDEX_EXP));
        SgExpression *index=exp->lhs()->rhs()->lhs()->lhs()->rhs();
        SgSymbol *scalarsym = index->symbol();
        SgSymbol * step0 = outer->getSymbolInCurrentScope(mixstring, "step", scalarsym->identifier(),0);
        SgExpression *step0exp = new SgVarRefExp(step0);
        exp->replace(*step0exp);
  }

}

//to optimize pcrc_glob() in indo, we have to find the innermost indo first
//then retrieve outer indo level by level
SgStatement * find_inner_indo(SgStatement * s, int level)
{
   int done = 0;
   SgStatement * stmt;

   switch(s->variant()) {
   case FOR_NODE: {
        indo_level++;

        int numberOfChildren=s->numberOfChildrenList1();
        SgStatement * * children = new SgStatement * [numberOfChildren];
        int i;

        for (i=0; i<numberOfChildren; i++)
                children[i]=s->childList1(i);

        for (i=0; i<numberOfChildren; i++) {
		if( (stmt = find_inner_indo(children[i], level)) != NULL){
			done = 1;
			break;
		}
	}

        delete [] children;

	indo_level--;

	if (done) return stmt;

        break;
        }

   default: {
	if (indo_level == level)
		return s->controlParent();
        break;
  	}
   }//end switch

   return NULL;

}


/*=================================================================
//First scan: 
//If there are stmts inside indo we can't deal with, treat it as DO,
//else, treat it as INDEPENDENT DO
//return: 0---DO, 1---INDEPENDENT DO
=================================================================*/
int scan_indo(SgStatement * s)
{
   int dec_indo = 0;

   switch(s->variant()) {
   case ASSIGN_STAT:
	break;

   case PROC_STAT: {
	int counter = 0;
        SgCallStmt * call = (SgCallStmt *)s;
        int number_Of_Args = call->numberOfArgs();
        SgFunctionSymb * prochedr=(SgFunctionSymb *)call->name();

        for(int i=0; i<number_Of_Args; i++){
        	if ((prochedr->parameter(i)->attributes() & OUT_BIT) || (prochedr->parameter(i)->attributes() & INOUT_BIT)) 
			counter++;
         };

	if (counter>=2) return 0;	//more than 2 out,not todo

        break;
	}
  
  case FORALL_STMT:
      	break;

  case INDEPENDENT_STMT:
	in_independent = on;
      	break;

  case FOR_NODE: {
        if (in_independent == on) {
           if(nindo_level>indo_level)  {
		//NOTE: for now, we assume there are no non-independent do
		//between two independent do! Else the semantec is not clear
                printf("independent do inside non_independent do!\n");
                return 0;
           }

           indo_level++;
           nindo_level++;

           in_independent = off;
	   dec_indo = 1;
        }
        else
                nindo_level++;

        int numberOfChildren=s->numberOfChildrenList1();
        SgStatement * * children = new SgStatement * [numberOfChildren];
        int i;

        for (i=0; i<numberOfChildren; i++)
                children[i]=s->childList1(i);

        for (i=0; i<numberOfChildren; i++)
                if (scan_indo(children[i]) == 0) 
			return 0;

        delete [] children;

        if (dec_indo) {
                indo_level--;
                nindo_level--;
        }
        else
                nindo_level--;

        break;
        }

  case ELSEIF_NODE:
  case LOGIF_NODE:
  case LOGIF_STMT: {
        int numberOfChildren1=s->numberOfChildrenList1();
        SgStatement * * children1 = new SgStatement * [numberOfChildren1];

        int numberOfChildren2=s->numberOfChildrenList2();
        SgStatement * * children2 = new SgStatement * [numberOfChildren2];

        int i;

        for (i=0; i<numberOfChildren1; i++)
                children1[i]=s->childList1(i);

        for (i=0; i<numberOfChildren1; i++)
                if (scan_indo(children1[i]) == 0) 
			return 0;

        delete [] children1;

        for (i=0; i<numberOfChildren2; i++)
                children2[i]=s->childList2(i);

        for (i=0; i<numberOfChildren2; i++)
                if (scan_indo(children2[i]) == 0) 
			return 0;

        delete [] children2;

        break;
        }

  case COMGOTO_NODE:
  case GOTO_NODE:
  case CONTINUE_NODE:
        break;

  case CONTROL_END:
  case ENDDO_STMT:
        //do nothing
        break;

  default: {
	if (!silentFlag)
        printf("Unknown statement encountered in indo: variant:%d\n",s->variant());
	return 0;
        break;
  	}
  }//end switch

  return 1;

}


/*=================================================================
 processing independent do header info
=================================================================*/
//fill indo header structure: process one layer of do each time
void fill_indo_header(Independent_header_def * indohdef, SgStatement * s)
{

  indohdef->setindexsymb(s->symbol());

  SgExpression *str = s->expr(0)->operand(2);   //s->expr(0): DDOT,
  if (str!=NULL)
    indohdef->setindextriple(2, str);           //step

  SgExpression *l =  s->expr(0)->operand(1)->operand(1);
  indohdef->setindextriple(0, l);
  SgExpression *u =  s->expr(0)->operand(1)->operand(2);
  indohdef->setindextriple(1, u);

}

//Second scan:
//find indo header recursively
void add_indo_header(Independent_header_def * indohdef, SgStatement * s)
{
   switch(s->variant()) {
   case INDEPENDENT_STMT:
        in_independent = on;
        break;

   case FOR_NODE: {
        if(in_independent == on) {
                fill_indo_header(indohdef, s);
                in_independent = off;

                int numberOfChildren=s->numberOfChildrenList1();
                SgStatement * * children = new SgStatement * [numberOfChildren];
                int i;

                for (i=0; i<numberOfChildren; i++)
                        children[i]=s->childList1(i);

                for (i=0; i<numberOfChildren; i++)
                        add_indo_header(indohdef, children[i]);

                delete [] children;
        }//end if

        break;
        }

   default:
        break;

   }//end switch

}

//This subroutine will prepare attribute structures for indo header.
void analyse_indo_header(SgStatement * s)
{
  Independent_header_def * indohdef = new Independent_header_def;

  //fill in outer header first
  fill_indo_header(indohdef, s);

  //find nested indo header
  int numberOfChildren=s->numberOfChildrenList1();
  SgStatement * * children = new SgStatement * [numberOfChildren];
  int i;

  for (i=0; i<numberOfChildren; i++)
        children[i]=s->childList1(i);

  for (i=0; i<numberOfChildren; i++)
        add_indo_header(indohdef, children[i]);

  delete [] children;

  // add this attributes to independent do statement
  s->addAttribute(INDEPHDER_Def,(void *) indohdef, sizeof(Independent_header_def));

}


/*================================================================
the following part deal with independent reference getting:
prepare attribute structures for indo reference,
the result include all var_ref & array_ref
================================================================*/

//prepare reference attribute for references in one expr
void get_indo_expression(SgStatement * s, SgExpression * expr, Usedfor side, int * count)
{
  if(expr) {
    switch(expr->variant()) {
    case BOOL_VAL:
    case INT_VAL:
    case FLOAT_VAL:
    case DOUBLE_VAL:
      break;

    case COMPLEX_VAL: {			
	// add cmplx()
      SgSymbol * sym=s->getSymbolInCurrentScope("cmplx");
      SgFunctionRefExp * funExp = new SgFunctionRefExp(*sym);
      funExp->addArg(*expr->operand(1));
      funExp->addArg(*expr->operand(2));
      expr->replace(*funExp);
      get_indo_expression(s, funExp->arg(0), side, count);
      get_indo_expression(s, funExp->arg(1), side, count);
      break;
    }
    
    case ARRAY_REF:
    case VAR_REF: {
      	independent_reference * indref = new independent_reference;

      	indref->setExpr(expr);

      	indref->sameAs = indo_find_same_as(s,expr);	

	indref->setIntent(side);

	(*count)++;
	if(side == as_lhs) {
		indref->setLeftCount(*count);
	}
	else {
		indref->setRightCount(*count);
	}

	indref->setStmt(s);

	//addAttribute to THIS stmt
      s->addAttribute(Independent_Ref,(void *)indref, sizeof(independent_reference));
      break;
    };

    case MINUS_OP:
    case ADD_OP:  
    case MULT_OP:
    case DIV_OP:
    case EXP_OP:
    case MOD_OP:

    case AND_OP:
    case OR_OP:
    case EQV_OP:
    case NEQV_OP:
      
    case EQ_OP:
    case NOTEQL_OP:
    case LT_OP:
    case LTEQL_OP:
    case GT_OP: 
    case GTEQL_OP: {
      get_indo_expression(s, expr->operand(1), side, count);
      get_indo_expression(s, expr->operand(2), side, count);
      break;
    }

    case NOT_OP: {
      get_indo_expression(s, expr->operand(1), side, count);
      break;
    }

    case PROC_CALL:
    case FUNC_CALL:
    case INTRIN_FUNC_CALL: {
      	//change the index reference in argument to pcrc_glob()
	//look for refs?

      	SgFunctionRefExp * funExp = (SgFunctionRefExp *) expr;
      	Independent_header_def * indohdef = (Independent_header_def *) s->controlParent()->getAttribute(0, INDEPHDER_Def)->getAttributeData();

      	for (int idx=0; idx<indohdef->getindexnumber(); idx++) {

      	   SgSymbol *indexsymb=indohdef->getindexsymb(idx);
           for (int i=0; i<funExp->numberOfArgs(); i++) {
		SgExpression *exp;
		exp = funExp->arg(i)->IsSymbolInExpression(*indexsymb);
		if (exp)
		{
      		independent_reference * indref= new independent_reference();
      		indref->setExpr(exp);
		change_indo_scalar_reference(s->controlParent(),indref);
		}
      	   }//end for
	}//end for
      break;
    }


    default:
      if (!silentFlag)
      printf("In get_indo_expression, unexpected expression on line %d, id: %d, varaint: %d\n",s->lineNumber(), expr->id(),expr->variant());

    };
  };
}

//prepare reference attribute structures for references in one stmt
void get_indo_children_reference(SgStatement * indo, SgStatement * child, int * lc, int * rc)
{
  //right_count starts from  0 in each stmt
  //left_count starts from value in last stmt
  (*rc)=0;

  switch(child->variant()) {
  case ASSIGN_STAT:
	intrinsic2sub(child);
	if (child->variant() == ASSIGN_STAT ) {
    		get_indo_expression(child, child->expr(0), as_lhs, lc);
    		get_indo_expression(child, child->expr(1), as_rhs, rc);
    		break;
	}

  case PROC_STAT: {
	SgCallStmt * call = (SgCallStmt *)child;
	int number_Of_Args = call->numberOfArgs();
	
	for(int i=0; i<number_Of_Args; i++){
		SgFunctionSymb * prochedr=(SgFunctionSymb *)call->name();

		if ((prochedr->parameter(i)->attributes() & OUT_BIT) || (prochedr->parameter(i)->attributes() & INOUT_BIT)) {
			get_indo_expression(call,call->arg(i),as_lhs,lc);
		}
		else {	
			get_indo_expression(call,call->arg(i),as_rhs,rc);
		};
	};

        //optimize, lan, 6.7,1997
        if(optimize_indo_call(call) == 1) {
                call->name()->addAttribute(INDO_CALL_OPT, NULL, 0);
        }

        call->name()->addAttribute(INDEP_Def, NULL, 0);

	break;
	};

  case INDEPENDENT_STMT:
    	break;

  case FOR_NODE: {
      	int numberOfChildren=child->numberOfChildrenList1();
      	SgStatement * * children = new SgStatement * [numberOfChildren];
      	int i;

      	for (i=0; i<numberOfChildren; i++)
        	children[i]=child->childList1(i);

      	for (i=0; i<numberOfChildren; i++)
        	get_indo_children_reference(child,children[i],lc,rc);

      	delete [] children;

      	break;
	}

  case LOGIF_NODE:
  case ELSEIF_NODE:
  case LOGIF_STMT: {
	//condition
    	get_indo_expression(child, child->expr(0), as_rhs, rc);

      	int numberOfChildren1=child->numberOfChildrenList1();
      	SgStatement * * children1 = new SgStatement * [numberOfChildren1];

      	int numberOfChildren2=child->numberOfChildrenList2();
      	SgStatement * * children2 = new SgStatement * [numberOfChildren2];

      	int i;

      	for (i=0; i<numberOfChildren1; i++)
        	children1[i]=child->childList1(i);

      	for (i=0; i<numberOfChildren1; i++)
        	get_indo_children_reference(child,children1[i],lc,rc);

      	delete [] children1;

      	for (i=0; i<numberOfChildren2; i++)
        	children2[i]=child->childList2(i);

      	for (i=0; i<numberOfChildren2; i++)
        	get_indo_children_reference(child,children2[i],lc,rc);

      	delete [] children2;

    	break;
	}

  case CONTROL_END:
  case ENDDO_STMT:
	//do nothing
    	break;

  default:
        if (!silentFlag)
    	printf("Not implemented now! variant:%d\n",child->variant());
    	break;
  };

}

//prepare reference attribute structures for all references inside indo
void get_indo_reference(SgStatement * s)
{
      int org_numberOfChildren=s->numberOfChildrenList1();
      int i;
      int *lcnt = new int;
      int *rcnt = new int;

	(*lcnt)=0;
	(*rcnt)=0;

      SgStatement * * children = new SgStatement * [org_numberOfChildren];

      for (i=0; i<org_numberOfChildren; i++)
        children[i]=s->childList1(i);

      for (i=0; i<org_numberOfChildren; i++)
        { 
		get_indo_children_reference(s,children[i],lcnt,rcnt);
	}

	delete lcnt, rcnt;
}

/*================================================================
After we got indo reference, we process it: establishing relationship
of reference with indo index, adding index-related info, such as 
index_level, offset and stride:
.index_indo_reference
	.index_indo_children_reference
		.process_indo_reference
================================================================*/
//process references in child stmt
void process_indo_reference(SgStatement * s, SgStatement * child, int lc)
{
  Independent_header_def * indohedr=(Independent_header_def *)s->getAttribute(0, INDEPHDER_Def)->getAttributeData();

  int num=child->numberOfAttributes(Independent_Ref);

  for (int idx=0; idx<indohedr->getindexnumber(); idx++) {
     SgSymbol * index=indohedr->getindexsymb(idx);

     for (int i=0; i<num; i++) {

	independent_reference * indref=(independent_reference *) ((child->getAttribute(i, Independent_Ref))->getAttributeData());

      	int indexlevel;

      	SgExpression * indrexp=indref->getexpression();
      	switch(indrexp->variant()) {
      	case VAR_REF:
		{
        	indexlevel=getreflevel(index, indrexp);
        	indref->setIndex(0, idx, indexlevel);
		//note: 1.var_ref might be array_ref in call arg-list
		//2.var_ref also could be lhs with array in rhs
	    	if(indref->getIntent()==as_rhs) 
		   indref->setLeftCount(lc);
        	break;
		}
      case ARRAY_REF:
	    	{
        	SgArrayRefExp * arrayexp=(SgArrayRefExp *) indrexp;
        	SgSymbol * arraysym=arrayexp->symbol();

        	for (int dimI=0; dimI<arrayexp->numberOfSubscripts();dimI++) {
          	    indexlevel=getreflevel(index, arrayexp->subscript(dimI));
          	    indref->setIndex(dimI, idx, indexlevel);

          	    // if possible, try to get stride and offset;
          	    if (indexlevel==1) {
            	        SgExpression * stride = getlinerexp(1, index, arrayexp->subscript(dimI));
            	        SgExpression * offset = getlinerexp(0, index, arrayexp->subscript(dimI));
            	        indref->setstride(dimI, idx, stride);
            	        indref->setoffset(dimI, idx, offset);
          	    } //end if
                 } //end for dimI
	
		//if rhs array, detect comm: none or remap
	    	if(indref->getIntent()==as_rhs) {

		   indref->setLeftCount(lc);

        	   if (detect_indo_no_communication(child,indref,1) == 1)
		   {
                	indref->setCommunication(NONE);
		   }
        	   else {
                	indref->setCommunication(REMAP);
		   }
		}

        	break;
      	    };//end ARRAY_REF

      	    default:
      		if (!silentFlag)
        	printf("process_indo_reference:Expression not touched!\n");
        	break;
	    }; //end switch
        };//end for 
   }//end for

   //addAttribute array_ref to do header for define_bound,var_ref for glob etc
   for (int i=0; i<num; i++) {
	independent_reference * indref=(independent_reference *) ((child->getAttribute(i, Independent_Ref))->getAttributeData());
	s->addAttribute(Independent_Ref,(void *)indref, sizeof(independent_reference));
   }

}

//index_indo_reference for each stmt inside indo
void index_indo_children_reference(SgStatement * s, SgStatement * child)
{
  switch(child->variant()) {
  case ASSIGN_STAT:
  case PROC_STAT:     {
	independent_reference * lhsref = get_lhsref(child);

	int lc;
	if(lhsref != NULL) {
		lhsref->setRightCount(0);
		lc =lhsref->getLeftCount();
	}
	else {	//no lhsref in this indo stmt, wrong!
      	    if (!silentFlag)
	    printf("index_indo_ref: Statement has no lhs:%d\n",child->id());
	    return;
	}

	process_indo_reference(s, child, lc);

	break;
     };//end PROC_STAT & ASSIGN_STAT

  case FOR_NODE: {
      	int numberOfChildren=child->numberOfChildrenList1();
      	SgStatement * * children = new SgStatement * [numberOfChildren];
      	int i;

      	for (i=0; i<numberOfChildren; i++)
        	children[i]=child->childList1(i);

      	for (i=0; i<numberOfChildren; i++)
        	index_indo_children_reference(s,children[i]);

      	delete [] children;

      	break;
	}

  case LOGIF_NODE:
  case LOGIF_STMT: {
	process_indo_reference(s, child, 0);

      	int numberOfChildren1=child->numberOfChildrenList1();
      	SgStatement * * children1 = new SgStatement * [numberOfChildren1];

      	int numberOfChildren2=child->numberOfChildrenList2();
      	SgStatement * * children2 = new SgStatement * [numberOfChildren2];

      	int i;

      	for (i=0; i<numberOfChildren1; i++)
        	children1[i]=child->childList1(i);

      	for (i=0; i<numberOfChildren1; i++)
        	index_indo_children_reference(s,children1[i]);

      	delete [] children1;

      	for (i=0; i<numberOfChildren2; i++)
        	children2[i]=child->childList2(i);

      	for (i=0; i<numberOfChildren2; i++)
        	index_indo_children_reference(s,children2[i]);

      	delete [] children2;

    	break;
	}

  case CONTROL_END:
  case ENDDO_STMT:
    break;

  default:
    if (!silentFlag)
    printf("Statement not touched yet!!!\n");
    break;
  } //outer switch 

}


//relate indo index to refs
//get indexlevel, stride and offset of the index var
//for most stmts, the op is same(?), just care the references:var,array
void index_indo_reference(SgStatement * s)
{

      int org_numberOfChildren=s->numberOfChildrenList1();
      int i;

      SgStatement * * children = new SgStatement * [org_numberOfChildren];

      for (i=0; i<org_numberOfChildren; i++)
        children[i]=s->childList1(i);

      for (i=0; i<org_numberOfChildren; i++)
        index_indo_children_reference(s,children[i]);

}


/*================================================================
First step in transforming: change indo loop bound: from global to local
.define_indo_bounds
.change_indo_header
BTW, change index var in stmt from (now) local to global
.calc_glob_index
================================================================*/
// calculate all loop bounds according to lhs arrays
// because we assume same computation partitioning now, only need to compute
// local loop bound according to one distributed array
int define_indo_bounds(SgStatement * s)
{
    Independent_header_def * indhdef = (Independent_header_def *)s->getAttribute(0, INDEPHDER_Def)->getAttributeData();
    int lc = -1;

    for (int i=0; i<indhdef->getindexnumber(); i++) {

  	for (int j=0; j<s->numberOfAttributes(Independent_Ref); j++) {
    	   independent_reference * lhsref = (independent_reference *) ((s->getAttribute(j, Independent_Ref))->getAttributeData());

	   lc = lhsref->getLeftCount();

	   if(lhsref->getIntent() == as_lhs){
	     if(lhsref->getexpression()->variant() == ARRAY_REF) {
 	  	SgArrayRefExp * array = (SgArrayRefExp *) lhsref->getexpression();

		int dim;
      	        for (dim=0; dim<array->numberOfSubscripts(); dim++)
           	   if (lhsref->getIndex(dim, i)>0)  break;
      		if (dim == array->numberOfSubscripts()) continue;
			
		//prepare info for index i
		SgSymbol * index=indhdef->getindexsymb(i);
		SgExpression * lbound = &(indhdef->getindextriple(i, 0)->copy());
		SgExpression * ubound = &(indhdef->getindextriple(i, 1)->copy());
		SgExpression * step;
        	if (indhdef->getindextriple(i, 2)!=NULL) {
	   	   step = &(indhdef->getindextriple(i, 2)->copy());
		}
		else {
           	   step = new SgValueExp(1);
		}

		//generate rng for loop, as below
      		SgSymbol * fname = s->getSymbolInCurrentScope(mixstring, "new_range_loop", NULL);
      		SgFunctionRefExp * functionexpr = new SgFunctionRefExp (*fname);

	      	{
            	functionexpr->addArg(*lbound);
            	functionexpr->addArg(*ubound);
            	functionexpr->addArg(*step);

        	{
		//find corresponding dim in array to the index var
		int ardim;
        	for (ardim=0; ardim<array->numberOfSubscripts(); ardim++) 
            	    if (lhsref->getIndex(ardim, i)>0) break;

          	SgExpression * stride;

          	if (lhsref->getstride(ardim, i)!=NULL) {
            	  if (lhsref->getstride(ardim, i)->isInteger())
              	    stride = new SgValueExp(lhsref->getstride(ardim, i)->valueInteger());
            	  else
              	    stride = lhsref->getstride(ardim, i);
          	}
          	else
            	  stride = new SgValueExp(0);

          	functionexpr->addArg(*stride);


          	SgExpression * offset;
          	if (lhsref->getoffset(ardim, i)!=NULL) {
            	  if (lhsref->getoffset(ardim, i)->isInteger())
              	    offset = new SgValueExp(lhsref->getoffset(ardim, i)->valueInteger());
            	  else
              	    offset = lhsref->getoffset(ardim, i);
          	}
          	else 
            	  offset = new SgValueExp(0);

          	functionexpr->addArg(*offset);

        	}; //end stride & offset

        	SgSymbol * distsym = lhsref->getexpression()->symbol();
        	SgSymbol * distdad = s->getSymbolInCurrentScope(mixstring,"dad",distsym->identifier());
        	SgValueExp * dimention = new SgValueExp (dim+1);
        	SgExpression *funcexpr = make_range(s,distdad,dimention);

        	functionexpr->addArg(*funcexpr);

      	      	}//end '{'

      		SgSymbol * irng = s->getSymbolInCurrentScope(mixstring, "irg", index->identifier(), lc);
      		SgExpression * rngexp = new SgExpression (VAR_REF, NULL, NULL, irng);
      		SgAssignStmt * assign = new SgAssignStmt (*rngexp, *functionexpr);
      		s->insertStmtBefore(*assign, *s->controlParent());

      		SgSymbol * funname = s->getSymbolInCurrentScope(mixstring, "delete_range", NULL);
      		SgCallStmt * afuncall = new SgCallStmt(*funname);
      		afuncall->addArg(* rngexp);
      		s->insertStmtAfter(*afuncall, *s->controlParent());

      		delete functionexpr, rngexp, assign , afuncall;

		//loop bounds
    		{
      		SgSymbol * fname = s->getSymbolInCurrentScope(mixstring, "loop_bounds", NULL);
      		SgCallStmt * afuncall = new SgCallStmt(*fname);


      		SgSymbol * irng = s->getSymbolInCurrentScope(mixstring, "irg", index->identifier(), lc);
      		SgExpression * rngexp = new SgExpression (VAR_REF, NULL, NULL, irng);
      		afuncall->addArg(*rngexp);

      		//SgSymbol * index=indhdef->getindexsymb();

      		SgSymbol * lb = s->getSymbolInCurrentScope(mixstring, "lil", index->identifier(), lc);
      		SgExpression * lbexpr = new SgExpression(VAR_REF, NULL, NULL, lb);
      		afuncall->addArg(*lbexpr);
      		SgSymbol * ub = s->getSymbolInCurrentScope(mixstring, "liu", index->identifier(), lc);
      		SgExpression * ubexpr = new SgExpression(VAR_REF, NULL, NULL, ub);
      		afuncall->addArg(*ubexpr);
      		SgSymbol * step = s->getSymbolInCurrentScope(mixstring, "lis", index->identifier(), lc);
      		SgExpression * stepexpr = new SgExpression(VAR_REF, NULL, NULL, step);
      		afuncall->addArg(*stepexpr);

      		s->insertStmtBefore(*afuncall, *s->controlParent());

      		delete afuncall,rngexp,lbexpr,ubexpr,stepexpr;

   		};

      		delete lbound,ubound,step;

		//only one loop bound is enough, lan 4/29/98
		break;
	    }
	    else { //special: var_ref, do nothing???

	    }//end if ARRAY_REF

	   }//end if as_lhs

	}//end for ref

   }//end for index i

	if (lc != -1) 
		return lc;
	else
		return -1;
}

//because of same computation partitioning, we change loop bound according to 
//the first distributed array
//the commented part might be needed later...
void change_indo_header(SgStatement * outer, SgStatement * s, int idx, int lc)
{
   	Independent_header_def * indhdef = (Independent_header_def *)outer->getAttribute(0, INDEPHDER_Def)->getAttributeData();
   	SgSymbol *index=indhdef->getindexsymb(idx);

	//int divided=0;	//indo loop to be divided?

        SgExpression *start,*end,*step; //original do_loop l,u,s
  	SgExpression *lbexpr,*ubexpr,*stepexpr;

	//independent_reference * lhsref;
/*
  	for (int i=0; i<outer->numberOfAttributes(Independent_Ref); i++) {
    	   lhsref = (independent_reference *) ((outer->getAttribute(i, Independent_Ref))->getAttributeData());

	   if(lhsref->getIntent() == as_lhs){
	     if(lhsref->getexpression()->variant() == ARRAY_REF) {
     		if (lhsref->refered(idx) != 0) {
			divided = 1;
			break;
		}
	     }
	   }
	}//end for

	if(divided){	//then change indo header, else leave it untouched

		int lc = lhsref->getLeftCount();
*/

     		lbexpr = new SgValueExp(0);

     		SgSymbol * ub = s->getSymbolInCurrentScope(mixstring, "liu", index->identifier(), lc);
     		ubexpr = new SgVarRefExp(ub);
     		{
       		SgSymbol * lb = s->getSymbolInCurrentScope(mixstring, "lil", index->identifier(), lc);
       		SgExpression * lbexpr = new SgVarRefExp(lb);
       		(*ubexpr) =(* ubexpr) - (* lbexpr);

       		SgSymbol * step = s->getSymbolInCurrentScope(mixstring, "lis", index->identifier(), lc);
       		SgExpression * stepexpr = new SgVarRefExp(step);
       		(*ubexpr) =(* ubexpr) / (* stepexpr);
     		}

     		stepexpr = new SgValueExp(1);

        	SgForStmt * forstmt=(SgForStmt *) s;
        	start = forstmt->start();
        	end = forstmt->end();
        	step = forstmt->step();

        	start->replace(*lbexpr);
        	end->replace(*ubexpr);
        	if (step!=NULL) step->replace(*stepexpr);

//	}// else do nothing

     	delete start,end,step,lbexpr,ubexpr,stepexpr;

}

//change local indo index to global index
//generate stmt: pcrc_glb_index= pcrc_glob(...)
void calc_glob_index(SgStatement * outer, SgStatement * s, int idx)
{
  Independent_header_def * indhdef = (Independent_header_def *)outer->getAttribute(0, INDEPHDER_Def)->getAttributeData();
  SgSymbol * index = indhdef->getindexsymb(idx);

  SgStatement * body = s->childList1(0);

  int num, dim;
  int change = 0;

  for (int i=0; i<outer->numberOfAttributes(Independent_Ref); i++) {
   	independent_reference * indref = (independent_reference *) ((outer->getAttribute(i, Independent_Ref))->getAttributeData());
  
	if(indref->getIntent() == as_lhs)
	   if(indref->getexpression()->variant() == ARRAY_REF) {
 	 	SgArrayRefExp * array = (SgArrayRefExp *) indref->getexpression();

		//see if this ARRAY_REF has irg
      	        for (dim=0;dim<array->numberOfSubscripts(); dim++)
           	   if (indref->getIndex(dim, idx)>0)  break;

      		if (dim != array->numberOfSubscripts()) {
			change = 1;
			num = i;
			break;
		}
	   }
  }//end for

  if(change) {
   	independent_reference * indref = (independent_reference *) ((outer->getAttribute(num, Independent_Ref))->getAttributeData());

	int lc = indref->getLeftCount();

      	SgSymbol * lg=s->getSymbolInCurrentScope(mixstring, "glob", NULL);
      	SgFunctionRefExp * lgfun = new SgFunctionRefExp(*lg);

      	SgSymbol * rng=s->getSymbolInCurrentScope(mixstring,"irg",index->identifier(), lc);
      	SgVarRefExp * rngexp= new SgVarRefExp(rng);

      	lgfun->addArg(*rngexp);

      	// we need to get local index;
      	SgExpression * indexexp= new SgVarRefExp(index);

      	SgSymbol * indexs = s->getSymbolInCurrentScope(mixstring, "lis", index->identifier(), lc);
      	SgExpression * indexsexp= new SgVarRefExp(indexs);
      	SgExpression * exp1 = &((* indexsexp) * (* indexexp));

      	SgSymbol * indexl = s->getSymbolInCurrentScope(mixstring, "lil", index->identifier(), lc);
      	SgExpression * indexlexp= new SgVarRefExp(indexl);
      	SgExpression * exp2 = &((* exp1) + (* indexlexp));

      	lgfun->addArg(*exp2);

      	SgSymbol * glb=s->getSymbolInCurrentScope(mixstring, "glb", index->identifier());
      	SgExpression * glbexp = new SgExpression (VAR_REF, NULL, NULL, glb);
      	SgAssignStmt * assign=new SgAssignStmt (*glbexp, *lgfun);

      	body->insertStmtBefore(*assign, *body->controlParent());

  }//end if
}


/*================================================================
Second step in transforming: change indo reference to local
(General toollike subroutines)
.create_indo_section
	.define_indo_section
.indo_add_coef
.indo_wrap_on
================================================================*/
//type: 0--call, 1--assignment
void define_indo_section(SgStatement * s, independent_reference * ref, int type)
{
  int idx=0;
  SgArrayRefExp * arrayexp = (SgArrayRefExp *) ref->getexpression();
  int lc=ref->getLeftCount();
  int rc=ref->getRightCount();

  int dimcounter=0;
  for (int i=0; i<arrayexp->numberOfSubscripts(); i++) {

    if( (type == 1) && ((idx=ref->indexNumberOnDim(i)) != -1)) { //distributed dimension

   	Independent_header_def * indohdef = (Independent_header_def *) s->getAttribute(0, INDEPHDER_Def)->getAttributeData();

        SgExpression * lbound = &(indohdef->getindextriple(idx, 0)->copy());
        SgExpression * ubound = &(indohdef->getindextriple(idx, 1)->copy());
        SgExpression * step;
        if (indohdef->getindextriple(idx, 2)!=NULL) {
           step = &(indohdef->getindextriple(idx, 2)->copy());
        }
        else {
           step = new SgValueExp(1);
        }

        //generate funcall
      SgSymbol * symb= s->getSymbolInCurrentScope(mixstring, "set_array_triplet", NULL);
      SgCallStmt * funcall= new SgCallStmt(*symb);

      {
        SgSymbol * symb = s->getSymbolInCurrentScope(mixstring, "sdd", NULL,lc,rc);
        SgExpression * dadsecexp = new SgExpression(VAR_REF, NULL, NULL, symb);
        funcall->addArg(*dadsecexp);
      }
      {
        dimcounter++;
        SgValueExp * secdim = new SgValueExp (dimcounter);
        funcall->addArg(*secdim);
      }
      {
          funcall->addArg(*lbound);
          funcall->addArg(*ubound);
          funcall->addArg(*step);
      }
      {
        SgSymbol * symb =s->getSymbolInCurrentScope(mixstring, "dad", arrayexp->symbol()->identifier());
        SgExpression * dadexp = new SgExpression(VAR_REF, NULL, NULL, symb);

        funcall->addArg(*dadexp);
      }
      {
        SgValueExp * dim = new SgValueExp (i+1);
        funcall->addArg(*dim);
      }

      s->insertStmtBefore(*funcall, *s->controlParent());

    }
    else {//ref in call, or not distributed

    if (!arrayexp->isScalarOnDim(i)) {
      SgSymbol * symb= s->getSymbolInCurrentScope(mixstring, "set_array_triplet", NULL);
      SgCallStmt * funcall= new SgCallStmt(*symb);

      {
        SgSymbol * symb = s->getSymbolInCurrentScope(mixstring, "sdd", NULL,lc,rc);
        SgExpression * dadsecexp = new SgExpression(VAR_REF, NULL, NULL, symb);

        funcall->addArg(*dadsecexp);
      }
      {
        dimcounter++;
        SgValueExp * secdim = new SgValueExp (dimcounter);
        funcall->addArg(*secdim);
      }
      {
        SgTripletExp * triplet=(SgTripletExp *)(arrayexp->subscript(i));
        for (int j=0; j<triplet->numberOfItems(); j++) {
          SgExpression * item = triplet->item(j);
        //  deal the form like a(:)
          if ( item == NULL) {
                SgSymbol * arrayname= arrayexp->symbol();
                SgExpression *subs;
                if ( j == 0)
                        subs = get_lbound(arrayname,i);
                if ( j == 1 )
                        subs = get_ubound(arrayname,i);
                funcall->addArg(*subs);
                continue;
          }
          item = calculateExp(item);
          funcall->addArg(*item);
        }       // end of for
        if (triplet->numberOfItems()<3) {
          SgValueExp * defaultstride = new SgValueExp(1);
          funcall->addArg(*defaultstride);
        }
      }
      {
        SgSymbol * symb =s->getSymbolInCurrentScope(mixstring, "dad", arrayexp->symbol()->identifier());
        SgExpression * dadexp = new SgExpression(VAR_REF, NULL, NULL, symb);

        funcall->addArg(*dadexp);
      }
      {
        SgValueExp * dim = new SgValueExp (i+1);
        funcall->addArg(*dim);
      }

      s->insertStmtBefore(*funcall, *s->controlParent());
    }
    else {
      SgSymbol * dadsymb = s->getSymbolInCurrentScope(mixstring, "sdd", NULL, lc, rc);
      SgExpression * scalar = arrayexp ->subscript(i);
      SgSymbol * dad =s->getSymbolInCurrentScope(mixstring, "dad", arrayexp->symbol()->identifier());
      int dim = i+1;
      SgStatement *funcall = make_set_array_scalar(s, dadsymb, scalar, dad, dim);
      s->insertStmtBefore(*funcall, *s->controlParent());
    }//end else
   }//end else
  }//end for

}

//type: 0--call, 1--assignment
void create_indo_section(SgStatement * s, independent_reference * ref, int type)
{
  	SgArrayRefExp * arrayexp = (SgArrayRefExp *) ref->getexpression();
  	int lc=ref->getLeftCount();
  	int rc=ref->getRightCount();

     	{	//new array section
  	SgSymbol * dadsec = s->getSymbolInCurrentScope(mixstring, "sdd", NULL, lc, rc);
  	SgSymbol * dad = s->getSymbolInCurrentScope(mixstring, "dad", arrayexp->symbol()->identifier());

  	int dim =0;
  	for (int i=0; i<arrayexp->numberOfSubscripts(); i++){
    	    if (!arrayexp->isScalarOnDim(i))
      		dim++;
	    if((type==1) && (ref->indexNumberOnDim(i) != -1))
			dim++;
	    
	}

  	SgStatement *assign= make_new_array_section(s, dadsec, dim, dad);

  	s->insertStmtBefore(*assign, *s->controlParent());
  	}

	//define section
	define_indo_section(s,ref, type);

	{	//delete section
  	SgSymbol * dadsec = s->getSymbolInCurrentScope(mixstring, "sdd", NULL, lc, rc);
  	SgStatement * funcall = make_delete_array(s , dadsec);
  	s->insertStmtAfter(*funcall, *s->controlParent());
	}
}


//to be a general add_coef subroutine
void indo_add_coef(SgStatement * s, SgArrayRefExp * array, int count)
{
  SgSymbol * arraysymb=array->symbol();

	//only need offset, no stride needed, so not "pcrc_coef"
  SgSymbol * coefsym=s->getSymbolInCurrentScope(mixstring, "offset", NULL);

  for (int i=0; i<array->numberOfSubscripts(); i++) {

    SgCallStmt * call = new SgCallStmt(*coefsym);

    {
      SgSymbol * dad = s->getSymbolInCurrentScope(mixstring, "dad", arraysymb->identifier());
      SgVarRefExp * dadexp = new SgVarRefExp(dad);
      call->addArg(*dadexp);
    };
    {
      SgValueExp * dimexp = new SgValueExp(i+1);
      call->addArg(*dimexp);
    }
    {
      SgExpression *dimexp = array->subscript(i);
      call->addArg(*dimexp);
    } 
    {
      	SgSymbol * indexsym; 
      	indexsym=s->getSymbolInCurrentScope(mixstring,"v",NULL,count,i);

      SgVarRefExp * indexexp = new SgVarRefExp(indexsym);
      call->addArg(*indexexp);
    }

	//pcrc_offset to be added before s later, not now! 
	//So wrap it as an attribute
    s->addAttribute(INDEX_INIT,call,sizeof(call));
  };	//end for

}


// add procossor position detection;
// wrap on for both assign and call stmt
void indo_wrap_on(SgStatement * s, int lc)
{

  	SgSymbol * lhssec = s->getSymbolInCurrentScope(mixstring, "sdd", NULL, lc, 0);
  	SgExpression * grpexp = make_group(s,lhssec);

  	SgExpression * onexp = make_on(s, grpexp);

  	SgStatement *newstmt = s->copyPtr();
  	SgStatement *ifonstmt= new SgIfStmt(*onexp, *newstmt);

  	for (int i=0; i<s->numberOfAttributes(INDEX_INIT); i++) {
           SgStatement *coef=(SgStatement *)s->attributeValue(i,INDEX_INIT);
           newstmt->insertStmtBefore(*coef, *newstmt->controlParent());
  	}

  	s->insertStmtAfter(*ifonstmt, *s->controlParent());
  	s->deleteStmt();
}


/*=====================================================================
Transform assign stmt inside indo
.trans_indo_assign
	.trans_indo_assign_reference
		.change_indo_assign_reference
=====================================================================*/
//note: assuming no communication at this time
//todo: if has comm, then maybe need stack
void change_indo_assign_reference(SgStatement * s, SgArrayRefExp * array, int count)
{
  	SgSymbol * arraysymb  = array->symbol();

      	SgSymbol * offsetsym = s->getSymbolInCurrentScope(mixstring, "v", NULL, count, 0);
	SgExpression * newexp = new SgVarRefExp(offsetsym);
	for(int i=1; i<array->numberOfSubscripts(); i++){
      		SgSymbol * offsetsym = s->getSymbolInCurrentScope(mixstring, "v", NULL, count, i);
		SgExpression * offsetexpr = new SgExpression(VAR_REF, NULL, NULL, offsetsym);
		(*newexp) = (* newexp) + (* offsetexpr);
	}//end for

	SgExpression * exp = newexp->copyPtr();

	array->subscripts()->replace(*exp);
// zheng.gb
	if (unknown_address(arraysymb))
        {
            change_array_by_stack(s, array);
        }

}


//based on comm attribute, creat SPMD comm stmts before indo and change 
//array references inside indo
void trans_indo_assign_reference(SgStatement * s, SgStatement * assign, int count)
{
   //for assign_stmt, ref no.0 is lhs
   independent_reference * lhsref=(independent_reference *) ((assign->getAttribute(0, Independent_Ref))->getAttributeData());
   independent_reference * indoref=(independent_reference *) ((assign->getAttribute(count, Independent_Ref))->getAttributeData());

   SgExpression * refexpr = (SgExpression *) indoref->getexpression();
   SgSymbol *symb = refexpr->symbol();

   Independent_header_def * indohdef = (Independent_header_def *) s->getAttribute(0, INDEPHDER_Def)->getAttributeData();
   for(int i=0; i<indohdef->getindexnumber(); i++){
      SgSymbol *indexsymb=indohdef->getindexsymb(i);

      //pcrc_glob() if needed
      SgExpression *exp;
      exp = refexpr->IsSymbolInExpression(*indexsymb);
      while (exp)	//note: previously if(exp) 
      {
   	independent_reference * tmpref= new independent_reference();
        tmpref->setExpr(exp);
        change_indo_scalar_reference(s,tmpref);

	//if still has index var, then do again
   	exp = refexpr->IsSymbolInExpression(*indexsymb);

      }//end while

   }//end for

   //todo: consider left var_ref case
   if (refexpr->variant()!=ARRAY_REF) return;

   SgArrayRefExp * array = (SgArrayRefExp *) refexpr;
   SgSymbol *arraysymb = array->symbol();

   if(indoref->getIntent()==as_lhs) {
	int lc = indoref->getLeftCount();
   	int need_sdd=1;	//lhs need dad? and need wrap on?

	//change reference of lhs
       	if ( find_collapse(assign,indoref) == 1 ) 
        	analyse_collapse(assign,array);      // calculate local subscripts directly
      	else {
		// add coef calls as attributes
        	indo_add_coef(assign, array, 0);  
        	change_indo_assign_reference(assign, array, 0);
	}

   }
   else {	//as_rhs

    printf("%s -- %s: ",lhsref->getexpression()->symbol()->identifier(),arraysymb->identifier());
    switch (indoref->getCommunication())
    {
    case NONE: {
        printf("rhs: none(scalar_assign)\n");
	int rc = indoref->getRightCount();
       	if ( find_collapse(assign,indoref) == 1 ) 
        	analyse_collapse(assign,array);   
      	else {
		// add coef calls as attributes
        	indo_add_coef(assign, array, rc);  
        	change_indo_assign_reference(assign, array, rc);
	}
        break;
       }
    case REMAP: {	//todo: later, now is not usable
        printf("remap(scalar_assign)\n");
        break;
        }
    }  // end switch

   }//end else

}

void trans_indo_assign(SgStatement * s, SgStatement * child)	
{
  SgAssignStmt * assign = (SgAssignStmt *) child;

  if (child->hasLabel()) {
    SgLabel * label = child->label();
    SgStatement * continueStmt = new SgStatement(CONTINUE_NODE);
    continueStmt->setLabel(*label);

    child->insertStmtBefore(*continueStmt,*child->controlParent());
    child->setLabel(NULL);
  }

  int i;

  for(i=0;i<assign->numberOfAttributes(Independent_Ref);i++){
	trans_indo_assign_reference(s,child,i);
  }

  for (i=0; i<assign->numberOfAttributes(Funccall_Ref); i++) {

        SgExpression *funcall = (SgExpression *)assign->attributeValue(i,Funccall_Ref);
        trans_funccall(assign,funcall);
  }

  //because assign_stmt, ref no.0 is lhs
  independent_reference * lhsref = (independent_reference *)assign->attributeValue(0, Independent_Ref);
  if (lhsref->getexpression()->variant()==ARRAY_REF ) {
     if(find_collapse(assign,lhsref) != 1) {//not collpased, need wrap on
        indo_wrap_on(assign, sec_number);
	return;
     }
  }

  for (i=0; i<assign->numberOfAttributes(INDEX_INIT); i++) {
      SgStatement *coef=(SgStatement *)assign->attributeValue(i,INDEX_INIT);
      assign->insertStmtBefore(*coef, *assign->controlParent());
  }

}

/*=====================================================================
Transform proc_call stmt inside indo
.trans_indo_procstmt
	.trans_indo_proc_reference
	.indo_change_procstmt
=====================================================================*/

void indo_change_procstmt(SgStatement * s, SgCallStmt * newcall, int opt)
{
  SgCallStmt * call = (SgCallStmt *)s;

  //find lhs_ref
  independent_reference * lhsref = get_lhsref(s);

  int lc; 
  if(lhsref != NULL)
  	lc =lhsref->getLeftCount();
  else {	//no lhsref in this indo stmt, wrong!
	printf("indo_change_proc: Statement has no lhs:%d\n",s->id());
	return;
  }

        //optimize
  if(opt) lc = sec_number;

  for (int i=0; i<call->numberOfArgs(); i++) {

    SgExpression * argexp = call->arg(i);

    switch(argexp->variant()) {
    case VAR_REF: {
        // if it is a scalar variable , leave it untouched !
      if ( argexp->type()->variant() != T_ARRAY ) {
        SgSymbol * argname = argexp->symbol();
        SgExpression * var = new SgExpression (VAR_REF, NULL, NULL, argname);
        newcall->addArg(*var);

        if (i==0 && (call->name()->attributes() & INTRINSIC_FUNC_BIT) )
        {
                SgSymbol *dadsymb=s->getSymbolInCurrentScope(mixstring, "dad", call->arg(i)->symbol()->identifier());
                SgExpression * dadexp = new SgExpression (VAR_REF, NULL, NULL, dadsymb);
                newcall->addArg(*dadexp);
        }
        break;
      }
        // else continue next step !
    };

    case ARRAY_REF: {
      SgArrayRefExp * arrayexp = (SgArrayRefExp *) argexp;
      SgSymbol * argname = s->getSymbolInCurrentScope(argexp->symbol()->identifier());

        // if is scalar array refernce , create stack to store it
      if (isSgArrayType(arrayexp->type()) == NULL  && argexp->numberOfAttributes(DUMMY_ADDR) == 0)
      {
        SgSymbol *dadsymb=s->getSymbolInCurrentScope(mixstring,"sdd",NULL,lc, 0);
        analyse_array_ref(s,arrayexp,s,dadsymb);
        newcall->addArg(*arrayexp);
        break;
      }

      SgArrayRefExp * var;

      if (argexp->numberOfAttributes(DUMMY_ADDR) != 0 )
        var = (SgArrayRefExp *)argexp->attributeValue(0,DUMMY_ADDR);
      else {
        var = new SgArrayRefExp(*argname);
        SgExpression *address = get_first_address(s,arrayexp);
        var->addSubscript(*address);
// zheng.gb
	if (unknown_address(argname))
        {
            change_array_by_stack(s, var);
        }
      }

      newcall->addArg(*var);

      SgSymbol *dadsymb;
      if (argexp->numberOfAttributes(DUMMY_NAME) != 0 )
                dadsymb = (SgSymbol *)argexp->attributeValue(0,DUMMY_NAME);
      else if (argexp->variant()==VAR_REF)
        dadsymb=s->getSymbolInCurrentScope(mixstring, "dad", call->arg(i)->symbol()->identifier());
      else
        dadsymb=s->getSymbolInCurrentScope(mixstring, "sdd", NULL, lc, 0);
      SgExpression * dadexp = new SgVarRefExp (dadsymb);
      newcall->addArg(*dadexp);

      break;
    };

    case INT_VAL:
    case FLOAT_VAL:
    case DOUBLE_VAL:
    case BOOL_VAL:
    case CHAR_VAL:
    case STRING_VAL:
    case COMPLEX_VAL: {
      // * do nothing, just copy the old one;
      newcall->addArg(*argexp);
      break;
    };

    default:
      // must be an scalar expression
      if (isSgArrayType(argexp->type()) == NULL )
      {
//printf("change_procstmt: expression encountered . \n");
        analyse_array_ref_in_exp(s,argexp,s);
        newcall->addArg(*argexp);
      }
      else
        newcall->addArg(*argexp);
      break;
    };

  }
  s->insertStmtAfter(*newcall, *s->controlParent());

}


void trans_indo_proc_reference(SgStatement * indo, SgStatement * s, int count)
{

  //this indref
  SgCallStmt * call = (SgCallStmt *)s;
  independent_reference * indref = (independent_reference *)call->getAttribute(count, Independent_Ref)->getAttributeData();

  //find lhsref
  independent_reference * lhsref = get_lhsref(s);

  int lc; 
  if(lhsref != NULL)
  	int lc = lhsref->getLeftCount();
  else {	//no lhsref in this indo stmt, wrong!
	printf("trans_indo_proc_reference: Statement has no lhs:%d\n",s->id());
	return;
  }

  SgExpression * var = indref->getexpression();

   Independent_header_def * indohdef = (Independent_header_def *) indo->getAttribute(0, INDEPHDER_Def)->getAttributeData();
   for(int i=0; i<indohdef->getindexnumber(); i++){
      	SgSymbol *indexsymb=indohdef->getindexsymb(i);

  	//change indo index reference
  	SgExpression *exp;
  	exp = var->IsSymbolInExpression(*indexsymb);
  	while (exp)
  	{
       	   independent_reference * tmpref = new independent_reference();
           tmpref->setExpr(exp);
           change_indo_scalar_reference(indo, tmpref);

           exp = var->IsSymbolInExpression(*indexsymb);
  	}//end while

   }//end for

   if (var->variant()!=ARRAY_REF) return;

   SgArrayRefExp * array = (SgArrayRefExp *) var;
   SgSymbol *arraysymb = array->symbol();

   if(indref->getIntent()==as_lhs) {
        //optimized
        if(call->name()->numberOfAttributes(INDO_CALL_OPT) == 0) {
                //different from assign, create section anyway
                create_indo_section(call, indref, 0);
        }
   }
   else {	//as_rhs
    	printf("%s -- %s: ",lhsref->getexpression()->symbol()->identifier(),arraysymb->identifier());
    	switch (indref->getCommunication()) {
    	case NONE: {
        	printf("indo call: none communication\n");
		//todo: ...
        	break;
       		}
    	case REMAP: {       
        	printf("indo call: remap communication\n");
        	printf("remap not implemented yet.\n");
		//todo: ...
        	break;
      	    	}
      	}  // end switch

   }//end if-else

}

void trans_indo_procstmt(SgStatement * indo, SgStatement * s)
{
  	SgCallStmt * call = (SgCallStmt *)s;
  	SgCallStmt * newcall = new SgCallStmt(*call->name());
	int opt_call = 0;

  	if (s->hasLabel()) {
    		SgLabel * label = s->label();
    		SgStatement * continueStmt = new SgStatement(CONTINUE_NODE);
    		continueStmt->setLabel(*label);

    		s->insertStmtBefore(*continueStmt,*s->controlParent());
    		s->setLabel(NULL);
  	}

  	int i;

  	for(i=0;i<s->numberOfAttributes(Independent_Ref);i++){
        	trans_indo_proc_reference(indo,s,i);
  	}

        if(call->name()->numberOfAttributes(INDO_CALL_OPT) != 0)
                opt_call = 1;

	//change proc_stmt references
        indo_change_procstmt(s, newcall, opt_call);

  	//find lhsref
  	independent_reference * lhsref = get_lhsref(s);

  	if(lhsref != NULL) {
            if (lhsref->getexpression()->variant()==ARRAY_REF) {
                if (opt_call) {
                        indo_wrap_on(newcall,sec_number);
                }
                else {
                        indo_wrap_on(newcall,lhsref->getLeftCount());
                }
           }
  	}
  	else {	//no lhsref in this indo stmt, wrong!
	    printf("trans_indo_procstmt: Statement has no lhs:%d\n",s->id());
  	}
}


/*=====================================================================
Transform IF stmt inside indo
.trans_indo_if_stmt
	.trans_indo_if_reference
=====================================================================*/
//change reference in if-condition, only as_rhs here
void trans_indo_if_reference(SgStatement * s, SgStatement * ifstmt, int count)
{
   independent_reference * indoref=(independent_reference *) ((ifstmt->getAttribute(count, Independent_Ref))->getAttributeData());

   SgExpression * refexpr = (SgExpression *) indoref->getexpression();
   SgSymbol *symb = refexpr->symbol();

   Independent_header_def * indohdef = (Independent_header_def *) s->getAttribute(0, INDEPHDER_Def)->getAttributeData();
   for(int i=0; i<indohdef->getindexnumber(); i++){
      	SgSymbol *indexsymb=indohdef->getindexsymb(i);

   	//pcrc_glob() if needed
   	SgExpression *exp;
   	exp = refexpr->IsSymbolInExpression(*indexsymb);
   	while (exp) 
   	{
           independent_reference * tmpref= new independent_reference();
           tmpref->setExpr(exp);
           change_indo_scalar_reference(s,tmpref);

           exp = refexpr->IsSymbolInExpression(*indexsymb);

   	}//end while

   }//end for

   if (refexpr->variant()!=ARRAY_REF) return;

   SgArrayRefExp * array = (SgArrayRefExp *) refexpr;

   //as_rhs
    switch (indoref->getCommunication())
    {
    case NONE: {
        printf("none(scalar_assign)\n");
        int rc = indoref->getRightCount();
        if ( find_collapse(ifstmt,indoref) == 1 )
                analyse_collapse(ifstmt,array);
        else {
                // add coef calls as attributes
		ifstmt->deleteAllAttribute(INDEX_INIT);
                indo_add_coef(ifstmt, array, rc);
                change_indo_assign_reference(ifstmt, array, rc);

  		for(int i=0;i<ifstmt->numberOfAttributes(INDEX_INIT);i++) 
		{
           	   SgStatement *coef=(SgStatement *)ifstmt->attributeValue(i,INDEX_INIT);
           	   ifstmt->insertStmtBefore(*coef, *ifstmt->controlParent());
  		}
        }

        break;
       }
    case REMAP: {       //todo: later, now is not usable
        printf("remap not implemented yet.\n");
        break;
        }
    }  // end switch

}

void trans_indo_if_stmt(SgStatement * indo, SgStatement * s)
{

  if (s->hasLabel()) {
    SgLabel * label = s->label();
    SgStatement * continueStmt = new SgStatement(CONTINUE_NODE);
    continueStmt->setLabel(*label);

    s->insertStmtBefore(*continueStmt,*s->controlParent());
    s->setLabel(NULL);
  }

  int i;

  for(i=0;i<s->numberOfAttributes(Independent_Ref);i++){
	trans_indo_if_reference(indo,s,i);
  }

}


/*=====================================================================
Framework of Indo transformation
.trans_indo
	.prepare_indo_section	//get "standard distributed array" info
	.trans_indo_children
=====================================================================*/
//process each stmt inside indo
void trans_indo_children(SgStatement * s, SgStatement * child, int lcount)
{
  switch(child->variant()) {
  case ASSIGN_STAT:
	trans_indo_assign(s, child);	
    	break;

  case PROC_STAT: 
	trans_indo_procstmt(s, child);	
    	child->deleteStmt();
    	break;

  case INDEPENDENT_STMT:
	in_independent = on;
    	child->deleteStmt();
    	break;

  case FOR_NODE: 
	{
      	int org_numberOfChildren=child->numberOfChildrenList1();
      	SgStatement * * children = new SgStatement * [org_numberOfChildren];
      	int i;
      	for (i=0; i<org_numberOfChildren; i++)
        	children[i]=child->childList1(i);

        if (in_independent == on) {//nested indo

	   indo_level++;
	   in_independent = off;
	   change_indo_header(s, child, indo_level, lcount);

           for (i=0; i<org_numberOfChildren; i++) {
		if(i==0) calc_glob_index(s, child, indo_level);
                trans_indo_children(s,children[i], lcount);
	   }

	   indo_level--;

	}
	else {	//common do loop

           for (i=0; i<org_numberOfChildren; i++)
                trans_indo_children(s,children[i], lcount);

	}//end if-else

      	delete [] children;

      	break;
    	}

  case LOGIF_NODE:
  case ELSEIF_NODE:
  case LOGIF_STMT: {
	trans_indo_if_stmt(s, child);	

      	int numberOfChildren1=child->numberOfChildrenList1();
      	SgStatement * * children1 = new SgStatement * [numberOfChildren1];

      	int numberOfChildren2=child->numberOfChildrenList2();
      	SgStatement * * children2 = new SgStatement * [numberOfChildren2];

      	int i;

      	for (i=0; i<numberOfChildren1; i++)
        	children1[i]=child->childList1(i);

      	for (i=0; i<numberOfChildren1; i++)
                trans_indo_children(s,children1[i], lcount);

      	delete [] children1;

      	for (i=0; i<numberOfChildren2; i++)
        	children2[i]=child->childList2(i);

      	for (i=0; i<numberOfChildren2; i++)
                trans_indo_children(s,children2[i], lcount);

      	delete [] children2;

    	break;
	}

  case CONTROL_END:
  case ENDDO_STMT:
	//do nothing
    	break;

  case COMGOTO_NODE:
  case GOTO_NODE:
  case CONTINUE_NODE:
        break;

  default:
    	printf("Not implemented now! variant:%d\n",s->variant());
    	break;

  };//end switch

}

int prepare_indo_section(SgStatement * s)
{
   independent_reference * lhsref;
   SgExpression * refexpr;
   SgSymbol *symb;

   Independent_header_def * indohdef = (Independent_header_def *) s->getAttribute(0, INDEPHDER_Def)->getAttributeData();

   //find indo reference for section
   int found = 0;
   for (int i=0; i<s->numberOfAttributes(Independent_Ref); i++) {
        lhsref=(independent_reference *) ((s->getAttribute(i, Independent_Ref))->getAttributeData());
        refexpr = (SgExpression *) lhsref->getexpression();
        symb = refexpr->symbol();

        //need to be array at lhs of assignment, and have distrinuted dimension
        if(lhsref->getIntent()!=as_lhs) continue;
        if (refexpr->variant()!=ARRAY_REF) continue;
	
        for(int j=0; j<indohdef->getindexnumber(); j++){
                if(lhsref->refered(j)) {
                        found = 1;
                        break;
                }
        }

        if(found) break;
   }

   //create section, for assignments in indo
   if(found) {
	create_indo_section(s, lhsref, 1);
	return lhsref->getLeftCount();
   }

	return -1;

}

void trans_indo(SgStatement * s)
{
   	Independent_header_def * indohdef = (Independent_header_def *) s->getAttribute(0, INDEPHDER_Def)->getAttributeData();
	SgStatement * inner_scope;
	int lcount;

	//divide computation, get local bounds, change indo header
	lcount = define_indo_bounds(s);

	if(lcount != -1) {
        	//prepare standard distributed array section, for on_proc
        	sec_number = prepare_indo_section(s);

		//change indo header, no add if(on)
		change_indo_header(s, s, indo_level, lcount);

      		int org_numberOfChildren=s->numberOfChildrenList1();
      		SgStatement * * children = new SgStatement * [org_numberOfChildren];
      		int i;

      		for (i=0; i<org_numberOfChildren; i++)
        		children[i]=s->childList1(i);

      		for (i=0; i<org_numberOfChildren; i++) {
			if(i==0) calc_glob_index(s, s, indo_level);
                	trans_indo_children(s,children[i], lcount);
		}

		//optimization
  	   if ( opt_strategy & LOCAL2GLOBAL ) {
		//find innermost indo
           	indo_level = 0;
		inner_scope = find_inner_indo(s,indohdef->getindexnumber());
        	indo_optimize_glob_index(s, inner_scope, indohdef, lcount);
	   }

      		delete [] children;

	}//end if lcount

}


/*=====================================================================
The following deal with subroutine body inside indo
.restrict_indo_call	
.trans_indo_call
	.add_coef_for_indo_proc
		.change_array_arg_in_expr
			.add_coef_for_expr
=====================================================================*/
//purge unhandleable syntax, might need changes in the future?
void restrict_indo_call(SgStatement * s)
{

      	int org_numberOfChildren=s->numberOfChildrenList1();
      	SgStatement * * children = new SgStatement * [org_numberOfChildren];

	int i;

      	for (i=0; i<org_numberOfChildren; i++)
        	children[i]=s->childList1(i);

      	for (i=0; i<org_numberOfChildren; i++) {
                switch(children[i]->variant()) {
                case INTERFACE_STMT:
                case PROCESSORS_STMT:
                case TEMPLATE_STMT:
                case ALIGN_DECL:
                case DISTRIBUTE_DECL:
                case INHERIT_DECL:
                case OUTOFCORE_DECL:
                case SHADOW_STMT:
                case INDEPENDENT_STMT:
                case FORALL_STMT: 
    			//printf("Syntax in indep do proc Not Supported: line %d,varaint: %d\n", children[i]->lineNumber(),children[i]->variant());
                        children[i]->deleteStmt();
                        break;

		case VAR_DECL:
			delete_dummy_array(children[i]);
                        break;
			
                default:
                        break;
                }
        }

}


//two tasks: 1.add pcrc_offset before s, 2.change reference to a(pcrc_v00)
//call pcrc_offset(pcrc_dad_a,1,index,pcrc_v00)
//call pcrc_offset(pcrc_dad_a,2,jndex,pcrc_v01)
//               a(pcrc_v00+pcrc_v01) = tmp
void add_coef_for_expr(SgStatement * s, SgExpression * exp, int count)
{
	//task 1:
  SgSymbol * coefsym=s->getSymbolInCurrentScope(mixstring, "offset", NULL);

  SgArrayRefExp * array = (SgArrayRefExp *) exp;
  SgSymbol * arraysymb = array->symbol();

  int i;

  for (i=0; i<array->numberOfSubscripts(); i++) {

    SgCallStmt * call = new SgCallStmt(*coefsym);

    {
      SgSymbol * dad = s->getSymbolInCurrentScope(mixstring, "dad", arraysymb->identifier());
      SgVarRefExp * dadexp = new SgVarRefExp(dad);
      call->addArg(*dadexp);
    };

    {
      SgValueExp * dimexp = new SgValueExp(i+1);
      call->addArg(*dimexp);
    }

    {
      SgExpression *dimexp = array->subscript(i);
      call->addArg(*dimexp);
    } 

    {
      SgSymbol * indexsym = s->getSymbolInCurrentScope(mixstring, "v", NULL, count, i);
      SgVarRefExp * indexexp = new SgVarRefExp(indexsym);
      call->addArg(*indexexp);
    }

    s->insertStmtBefore(*call, *s->controlParent());
  };	//end for

	//task 2:
      	SgSymbol * offsetsym = s->getSymbolInCurrentScope(mixstring, "v", NULL, count, 0);
	SgExpression * newexp = new SgExpression(VAR_REF, NULL, NULL, offsetsym);
	for(i=1; i<array->numberOfSubscripts(); i++){
      		SgSymbol * offsetsym = s->getSymbolInCurrentScope(mixstring, "v", NULL, count, i);
		SgExpression * offsetexpr = new SgExpression(VAR_REF, NULL, NULL, offsetsym);
		(*newexp) = (* newexp) + (* offsetexpr);
	}

	array->subscripts()->replace(*newexp);
}

//change array expression to local reference inside indo subroutine body
int change_array_arg_in_expr(SgStatement * s, SgSymbol * arga, SgExpression * expr, int count, int opt)
{
  if(expr) {
    switch(expr->variant()) {
    case BOOL_VAL:
    case INT_VAL:
    case FLOAT_VAL:
    case DOUBLE_VAL:
    case VAR_REF: 
      break;

    case COMPLEX_VAL: {			
	// add cmplx()
      SgSymbol * sym=s->getSymbolInCurrentScope("cmplx");
      SgFunctionRefExp * funExp = new SgFunctionRefExp(*sym);
      funExp->addArg(*expr->operand(1));
      funExp->addArg(*expr->operand(2));
      expr->replace(*funExp);
	int newc = count;
	newc = change_array_arg_in_expr(s,arga,funExp->arg(0),newc,opt);
	count = change_array_arg_in_expr(s,arga,funExp->arg(1),newc,opt);
      	break;
    }
    
    case ARRAY_REF: {
	if(opt == 0) { //no optimization
	   if(expr->symbol() == arga) {
		count++;
		add_coef_for_expr(s,expr,count);
	   }
	}
	else {
	   SgArrayRefExp *arrayexp = (SgArrayRefExp *)expr;
	   analyse_collapse(s,arrayexp);
	}
	
      	break;
    };

    case MINUS_OP:
    case ADD_OP:  
    case MULT_OP:
    case DIV_OP:
    case EXP_OP:
    case MOD_OP:

    case AND_OP:
    case OR_OP:
    case EQV_OP:
    case NEQV_OP:
      
    case EQ_OP:
    case NOTEQL_OP:
    case LT_OP:
    case LTEQL_OP:
    case GT_OP: 
    case GTEQL_OP: {
	int newc = count;
	newc = change_array_arg_in_expr(s, arga, expr->operand(1), newc,opt);
	count = change_array_arg_in_expr(s, arga, expr->operand(2), newc,opt);
      	break;
    }

    case NOT_OP: {
	int newc = count;
	newc = change_array_arg_in_expr(s, arga, expr->operand(1), newc,opt);
	count = newc;
      	break;
    }

    case PROC_CALL:
    case FUNC_CALL:
    case INTRIN_FUNC_CALL: {
	int newc = count;
      	SgFunctionRefExp * funExp = (SgFunctionRefExp *) expr;
      	for (int i=0; i<funExp->numberOfArgs(); i++) {
		newc = change_array_arg_in_expr(s, arga, funExp->arg(0), newc,opt);
      	};
	count = newc;
    	}
    	break;

    default:
      	printf("In change_array_arg_in_expr, unexpected expression on line %d, id: %d, varaint: %d\n",s->lineNumber(), expr->id(),expr->variant());

    };
  };
	return count;
}

//for each stmt, find array expression to change
void add_coef_for_indo_proc(SgStatement * s, SgSymbol * arga, int opt)
{
	int count = 0 ;

	switch(s->variant()) {
	case PROC_HEDR:{
      		int org_numberOfChildren=s->numberOfChildrenList1();
      		SgStatement * * children = new SgStatement * [org_numberOfChildren];
		int i;
      		for (i=0; i<org_numberOfChildren; i++)
        		children[i]=s->childList1(i);

      		for (i=0; i<org_numberOfChildren; i++) 
			add_coef_for_indo_proc(children[i], arga, opt);
		}
		break;

	case VAR_DECL:	
		break;

	case RETURN_STAT:{
		if (s->expr(0) != NULL){
			int newc;
			newc = change_array_arg_in_expr(s, arga,s->expr(0), count,opt);
			count = newc;
			};
		}
		break;

	case PROC_STAT:{
		int newc = count;

		SgCallStmt * call = (SgCallStmt *)s;
		int number_Of_Args = call->numberOfArgs();
		for(int j=0; j<number_Of_Args; j++){
			newc = change_array_arg_in_expr(call, arga,call->arg(j), newc,opt);
		}; 
		count = newc;
		
		}
		break;

	case ASSIGN_STAT:{
		int newc = count;
		newc = change_array_arg_in_expr(s, arga,s->expr(0), newc,opt);
		newc = change_array_arg_in_expr(s, arga,s->expr(1), newc,opt);
		count = newc;
		}
		break;

/*
  	case LOGIF_STMT:{	//only one body stmt, really exists?
		int newc = count;
		SgLogIfStmt * logif = (SgLogIfStmt *)s;
		//check condition
		newc = change_array_arg_in_expr(logif, arga, logif->conditional(), newc,opt);
		count = newc;
			
		add_coef_for_indo_proc(logif->body(), arga, opt);

		}
		break;
*/

  	case LOGIF_STMT: {
		s->setVariant(LOGIF_NODE);
		}
  	case LOGIF_NODE:{ 
		int i;
		int newc = count;
		SgIfStmt * ifstmt = (SgIfStmt *)s;

		//check condition
		newc = change_array_arg_in_expr(ifstmt, arga, ifstmt->conditional(), newc,opt);
		count = newc;

		//check children: true
      		int numberOfChildren1=s->numberOfChildrenList1();
      		SgStatement * * children1 = new SgStatement * [numberOfChildren1];
      		for (i=0; i<numberOfChildren1; i++)
        		children1[i]=s->childList1(i);

      		for (i=0; i<numberOfChildren1; i++)
			add_coef_for_indo_proc(children1[i], arga, opt);

		//false
      		int numberOfChildren2=s->numberOfChildrenList2();
      		SgStatement * * children2 = new SgStatement * [numberOfChildren2];
      		for (i=0; i<numberOfChildren2; i++)
        		children2[i]=s->childList2(i);

      		for (i=0; i<numberOfChildren2; i++)
			add_coef_for_indo_proc(children2[i], arga, opt);
		}
		break;

  	case ELSEIF_NODE:{	
		int i;
		int newc = count;
		SgIfStmt * ifstmt;
		SgStatement * inner2;
		SgIfStmt * elseif = (SgIfStmt *)&s->copy();

		//change elsif to else { if ... 
		SgExpression * cond = &(elseif->conditional())->copy();
		SgStatement * inner1 = &(elseif->childList1(0))->copy();
		if(elseif->childList2(0)!=NULL) {
		  	inner2 = &(elseif->childList2(0))->copy();
  		  	ifstmt = new SgIfStmt(*cond, *inner1, *inner2);
		}
		else{
  		  ifstmt = new SgIfStmt(*cond, *inner1);
		}
		
		SgStatement * last = inner1;
		SgStatement * next;
      		int numberOfChildren1=elseif->numberOfChildrenList1();
      		for (i=1; i<numberOfChildren1; i++) {
			next = &(elseif->childList1(i))->copy();
			last->insertStmtAfter(*next,*ifstmt);
			last = next;
		}
		last = inner2;
      		int numberOfChildren2=elseif->numberOfChildrenList2();
      		for (i=1; i<numberOfChildren2; i++) {
			next = &(elseif->childList2(i))->copy();
			last->insertStmtAfter(*next,*ifstmt);
			last = next;
		}
		
      		s->insertStmtAfter(*ifstmt, *s->controlParent());
		s->deleteStmt();

		//check condition
		newc = change_array_arg_in_expr(ifstmt, arga, ifstmt->conditional(), newc,opt);
		count = newc;

		//check children: true
      		numberOfChildren1=ifstmt->numberOfChildrenList1();
      		SgStatement * * children1 = new SgStatement * [numberOfChildren1];
      		for (i=0; i<numberOfChildren1; i++)
        		children1[i]=ifstmt->childList1(i);

      		for (i=0; i<numberOfChildren1; i++)
			add_coef_for_indo_proc(children1[i], arga, opt);

		//false
      		numberOfChildren2=ifstmt->numberOfChildrenList2();
      		SgStatement * * children2 = new SgStatement * [numberOfChildren2];
      		for (i=0; i<numberOfChildren2; i++)
        		children2[i]=ifstmt->childList2(i);

      		for (i=0; i<numberOfChildren2; i++)
			add_coef_for_indo_proc(children2[i], arga, opt);
		}
		break;

	case DO_WHILE_NODE: {
		int newc = count;
		SgDoWhileStmt * dowhile = (SgDoWhileStmt *)s;
		//check condition
		newc = change_array_arg_in_expr(dowhile, arga, dowhile->conditional(), newc,opt);
		count = newc;

		//check children
      		int org_numberOfChildren=s->numberOfChildrenList1();
      		SgStatement * * children = new SgStatement * [org_numberOfChildren];
		int i;
      		for (i=0; i<org_numberOfChildren; i++)
        		children[i]=s->childList1(i);

      		for (i=0; i<org_numberOfChildren; i++) 
			add_coef_for_indo_proc(children[i], arga, opt);
		}
		break;

	case FOR_NODE:{	
		int newc = count;

		SgForStmt * loop = (SgForStmt *)s;

		//check triplet
		newc = change_array_arg_in_expr(loop, arga, loop->start(), count,opt);
		newc = change_array_arg_in_expr(loop, arga, loop->end(), newc,opt);
		if(loop->step() != NULL){
			newc = change_array_arg_in_expr(loop, arga, loop->step(), newc,opt);
		};
		count = newc;

		//check children
      		int org_numberOfChildren=s->numberOfChildrenList1();
      		SgStatement * * children = new SgStatement * [org_numberOfChildren];
		int i;
      		for (i=0; i<org_numberOfChildren; i++)
        		children[i]=s->childList1(i);

      		for (i=0; i<org_numberOfChildren; i++) 
			add_coef_for_indo_proc(children[i], arga, opt);
		}
		break;

  	case CONTROL_END:
  	case ENDDO_STMT:
    		break;

        default:
		printf("In add_coef, unexpected statement on line %d, id: %d, variant: %d\n", s->lineNumber(),s->id(), s->variant());
                break;
        }//end switch
}

void trans_indo_call(SgStatement * s)	//consider to devide this proc ? 
{					//first delete , then add_coef

	SgProcHedrStmt * prochedr = (SgProcHedrStmt *)s;
      	int org_numberOfChildren=s->numberOfChildrenList1();
      	SgStatement * * children = new SgStatement * [org_numberOfChildren];

	//optimize: simplify array reference
	int  opt_call = 0;
	if (prochedr->name().numberOfAttributes(INDO_CALL_OPT) > 0) {
		opt_call = 1;
	}

	//purify
	restrict_indo_call(s);

	SgSymbol * array_arg;	//later consider an array of array symbols
	int i;

	//find the array argument, for coef
	for (i=0; i<prochedr->numberOfParameters(); i++)
	{
		if(prochedr->parameter(i)->type()->variant() == T_ARRAY){
			printf("in indo coef: i=%d\n", i);
			array_arg = prochedr->parameter(i);

			//change array reference inside subroutine to coef
			add_coef_for_indo_proc(s,array_arg,opt_call); 
		}
	}

	change_prochedr(s);

}

/*=====================================================================
Analyse Indo: header info & reference indo
=====================================================================*/
//1. indo header
//2. all array references inside indo
void analyse_indo(SgStatement * s)
{
	analyse_indo_header(s);	//ok for now

	get_indo_reference(s);	//add other components

        index_indo_reference(s); //process indo_ref's

}

